@@ -35,6 +35,7 @@
 #include <stdarg.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <louis.h>
 #include "louisxml.h"
 #include "sem_names.h"
 
@@ -69,8 +70,7 @@ configureError (FileInfo * nested, char *format, ...)
 #endif
   va_end (arguments);
   if (nested)
-    lou_logPrint ("File %s line %d: %s",
-		  nested->fileName, nested->lineNumber, buffer);
+    lou_logPrint ("%s:%d: %s", nested->fileName, nested->lineNumber, buffer);
   else
     lou_logPrint ("%s", buffer);
   errorCount++;
@@ -368,18 +368,27 @@ ignoreCaseComp (const char *str1, const char *str2, int length)
   return 0;
 }
 
+int
+find_action (const char **actions, const char *action)
+{
+  int actionLength = strlen (action);
+  int k;
+  for (k = 0; actions[k]; k += 2)
+    if (actionLength == strlen (actions[k])
+	&& ignoreCaseComp (actions[k], action, actionLength) == 0)
+      break;
+  if (actions[k] == NULL)
+    return -1;
+  return atoi (actions[k + 1]);
+}
+
 static int
 checkActions (FileInfo * nested, const char **actions)
 {
-  int k;
-  for (k = 0; actions[k]; k += 2)
-    if (nested->actionLength == strlen (actions[k]) &&
-	ignoreCaseComp
-	(actions[k], nested->action, nested->actionLength) == 0)
-      break;
-  if (actions[k] == NULL)
+  int actionNum = find_action (actions, nested->action);
+  if (actionNum == -1)
     return NOTFOUND;
-  return atoi (actions[k + 1]);
+  return actionNum;
 }
 
 static int
@@ -504,48 +513,29 @@ compileConfig (FileInfo * nested)
     "33",
     "linefill",
     "34",
+    "debug",
+    "35",
     "style",
     "90",
     NULL
   };
   static const char *yesNo[] = {
-    "no",
-    "0",
-    "yes",
-    "1",
-    NULL
+    "no", "0", "yes", "1", NULL
   };
   static const char *topBottom[] = {
-    "bottom",
-    "0",
-    "top",
-    "1",
-    NULL
+    "bottom", "0", "top", "1", NULL
   };
   static const char *encodings[] = {
-    "utf8",
-    "0",
-    "utf16",
-    "1",
-    "utf32",
-    "2",
-    "ascii8",
-    "3",
-    NULL
+    "utf8", "0", "utf16", "1", "utf32", "2", "ascii8", "3", NULL
   };
   static const char *backFormats[] = {
-    "plain",
-    "0",
-    "html",
-    "1",
-    NULL
+    "plain", "0", "html", "1", NULL
   };
 
   static const char *formatFor[] = {
-    "textDevice",
-    "0",
-    "browser",
-    "1",
+    "textDevice", "0",
+    "browser", "1",
+    "utd", "2",
     NULL
   };
 
@@ -691,7 +681,9 @@ compileConfig (FileInfo * nested)
 	  break;
 	case 30:
 	  {
-	    static const char *actions[] = { NULL };
+	    static const char *actions[] = {
+	      NULL
+	    };
 	    if (nested->value == NULL)
 	      configureError (nested, "a file name in column 2 is required");
 	    else
@@ -720,6 +712,10 @@ compileConfig (FileInfo * nested)
 	  else
 	    ud->line_fill = nested->value[0];
 	  break;
+	case 35:
+	  if ((k = checkValues (nested, yesNo)) != NOTFOUND)
+	    ud->debug = k;
+	  break;
 	case 90:
 	  {
 	    static const char *actions[] = {
@@ -742,10 +738,7 @@ compileConfig (FileInfo * nested)
 	      "newPageAfter",
 	      "10",
 	      "rightHandPage",
-	      "11",
-	      "braillePageNumberFormat",
-	      "12",
-	      NULL
+	      "11", "braillePageNumberFormat", "12", NULL
 	    };
 	    static const char *formats[] = {
 	      "leftJustified",
@@ -761,23 +754,10 @@ compileConfig (FileInfo * nested)
 	      "listColumns",
 	      "5",
 	      "listLines",
-	      "6",
-	      "computerCoded",
-	      "7",
-	      "contents",
-	      "8",
-	      NULL
+	      "6", "computerCoded", "7", "contents", "8", NULL
 	    };
 	    static const char *pageNumFormats[] = {
-	      "normal",
-	      "0",
-	      "blank",
-	      "1",
-	      "p",
-	      "2",
-	      "roman",
-	      "3",
-	      NULL
+	      "normal", "0", "blank", "1", "p", "2", "roman", "3", NULL
 	    };
 	    StyleType *style;
 	    sem_act styleAction;
@@ -897,7 +877,7 @@ initConfigFiles (const char *firstConfigFile, char *fileName)
 }
 
 int
-read_configuration_file (const char * configFileList, const char
+read_configuration_file (const char *configFileList, const char
 			 *logFileName,
 			 const char *configString, unsigned int mode)
 {
@@ -909,7 +889,7 @@ read_configuration_file (const char * configFileList, const char
   int currentListPos = 0;
   errorCount = 0;
   logFileNamex = logFileName;
-  if (mode != 0)
+  if (mode & dontInit)
     {
       ud->has_comp_code = 0;
       ud->has_math = 0;
@@ -934,6 +914,7 @@ read_configuration_file (const char * configFileList, const char
     return 0;
   memset (ud, 0, sizeof (UserData));
   entities = 0;
+  ud->mode = mode;
   ud->top = -1;
   ud->style_top = -1;
   for (k = document; k < notranslate; k++)
@@ -995,9 +976,6 @@ read_configuration_file (const char * configFileList, const char
 	  memset (ud->typeform, 0, sizeof (ud->typeform));
 	}
     }
-
-  if (errorCount)
-    lou_logPrint ("%d errors found, trying to continue", errorCount);
   ud->braille_page_number = ud->beginning_braille_page_number;
   if (entities)
     strcat (ud->xml_header, "]>\n");
