@@ -14,51 +14,15 @@ static const char *semNames[] = {
   "configtweak",
   "document",
   "para",
-  "list",
-  "caption",
-  "exercise1",
-  "exercise2",
-  "exercise3",
-  "directions",
-  "stanza",
-  "line",
   "heading1",
   "heading2",
   "heading3",
   "heading4",
-  "section",
-  "subsection",
-  "table",
-  "volume",
-  "titlepage",
-  "dedication",
   "contentsheader",
   "contents1",
   "contents2",
   "contents3",
   "contents4",
-  "code",
-  "quotation",
-  "attribution",
-  "spatial",
-  "arith",
-  "matrix",
-  "note",
-  "trnote",
-  "dispmath",
-  "disptext",
-  "music",
-  "graph",
-  "graphlabel",
-  "indexx",
-  "glossary",
-  "biblio",
-  "blanklinebefore",
-  "style1",
-  "style2",
-  "style3",
-  "style4",
-  "style5",
   "notranslate",
   "compbrl",
   "uncontracted",
@@ -66,30 +30,16 @@ static const char *semNames[] = {
   "pagenum",
   "runninghead",
   "footer",
-  "tblhead",
-  "tblbody",
-  "tblrow",
-  "tblcol",
-  "frontmatter",
-  "acknowledge",
-  "bodymatter",
-  "rearmatter",
   "boxline",
   "italicx",
   "boldx",
-  "center",
+  "linespacing",
   "blankline",
   "softreturn",
   "newpage",
   "righthandpage",
-  "endnotes",
-  "preface",
-  "jacket",
-  "copyright",
-  "booktitle",
-  "author",
-  "transcriber",
-  "tnpage",
+  "code",
+  "music",
   "math",
   "mi",
   "mn",
