@@ -33,6 +33,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include "louisxml.h"
+#include <libxml/xpath.h>
 #include "sem_names.h"
 
 typedef struct
@@ -56,11 +57,18 @@ typedef struct
   widechar charInserts[MAXINSERTS];
 } InsertsType;
 
+typedef enum
+{
+  styleEntry = 1,
+  xpathEntry = 2
+} EntryType;
+
 typedef struct
 {
   void *next;
   unsigned char *key;
-  int value;
+  EntryType type;
+  int semNum;
   InsertsType *inserts;
   StyleType *style;
 } HashEntry;
@@ -76,8 +84,10 @@ static int notFound = -1;
 static HashTable *actionTable = NULL;
 static HashTable *semanticTable = NULL;
 static HashTable *newEntriesTable;
-
 static int errorCount = 0;
+static xmlXPathContext *xpathCtx = NULL;
+static int registerNamespaces (FileInfo * Nested, xmlXPathContextPtr
+			       xpathCtx, const xmlChar * nsList);
 
 static void
 semanticError (FileInfo * nested, char *format, ...)
@@ -92,8 +102,7 @@ semanticError (FileInfo * nested, char *format, ...)
 #endif
   va_end (arguments);
   if (nested)
-    lou_logPrint ("File %s line %d: %s",
-		  nested->fileName, nested->lineNumber, buffer);
+    lou_logPrint ("%s:%d: %s", nested->fileName, nested->lineNumber, buffer);
   else
     lou_logPrint ("%s", buffer);
   errorCount++;
@@ -140,7 +149,7 @@ hashFree (HashTable * table)
 	free (e->key);
 	if (e->inserts != NULL)
 	  free (e->inserts);
-	if (e->value == -100)
+	if (e->type & styleEntry)
 	  free (e->style);
 	free (e);
       }
@@ -150,6 +159,12 @@ hashFree (HashTable * table)
 static HashEntry *
 hashScan (HashTable * table)
 {
+/* This function should be called in a statement like
+* while ((curEnt=y = hashScan (tablename)))
+* where
+* HasEntry *curEntry;
+/ It returns a table entry at each call, in no order. When it reaches 
+* the end of the table it resets itself and returns NULL. */
   HashEntry *e;
   if (table == NULL)
     return NULL;
@@ -180,8 +195,8 @@ static HashEntry *latestEntry;
 
 /* assumes that key is not already present! */
 static void
-hashInsert (HashTable * table, const unsigned char *key, int value,
-	    InsertsType * inserts, StyleType * style)
+hashInsert (HashTable * table, const unsigned char *key, int type, int
+	    semNum, InsertsType * inserts, StyleType * style)
 {
   int i;
   if (table == NULL || key == NULL)
@@ -191,7 +206,8 @@ hashInsert (HashTable * table, const unsigned char *key, int value,
   latestEntry->next = table->entries[i];
   latestEntry->key = malloc (strlen ((char *) key) + 1);
   strcpy ((char *) latestEntry->key, (char *) key);
-  latestEntry->value = value;
+  latestEntry->type = type;
+  latestEntry->semNum = semNum;
   latestEntry->inserts = inserts;
   latestEntry->style = style;
   table->entries[i] = latestEntry;
@@ -218,12 +234,11 @@ hashLookup (HashTable * table, const unsigned char *key)
 	if (key[k] != latestEntry->key[k])
 	  break;
       if (k == keyLength)
-	return latestEntry->value;
+	return latestEntry->semNum;
     }
   return notFound;
 }
 
-static char oldFileList[MAXNAMELEN];
 static char firstFileName[MAXNAMELEN];
 static int haveSemanticFile = 1;
 static int docNewEntries = 1;
@@ -234,6 +249,7 @@ find_semantic_number (const char *name)
   static const char *pseudoActions[] = {
     "include",
     "newentries",
+    "namespaces",
     NULL
   };
   int k;
@@ -242,12 +258,12 @@ find_semantic_number (const char *name)
     {
       actionTable = hashNew ();
       for (k = 0; k < end_all; k++)
-	hashInsert (actionTable, (xmlChar *) semNames[k], k, NULL, NULL);
+	hashInsert (actionTable, (xmlChar *) semNames[k], 0, k, NULL, NULL);
       k = 0;
       while (pseudoActions[k] != NULL)
 	{
 	  hashInsert (actionTable, (xmlChar *) pseudoActions[k],
-		      k + end_all + 1, NULL, NULL);
+		      0, k + end_all + 1, NULL, NULL);
 	  k++;
 	}
     }
@@ -266,6 +282,9 @@ destroy_semantic_table (void)
   newEntriesTable = NULL;
   hashFree (actionTable);
   actionTable = NULL;
+  if (xpathCtx != NULL)
+    xmlXPathFreeContext (xpathCtx);
+  xpathCtx = NULL;
 }
 
 static widechar
@@ -312,7 +331,7 @@ encodeInsertions (FileInfo * nested, xmlChar * insertions, int length)
   inserts.lastInsert = 0;
   k = 0;
   prevk = 0;
-/*Inserjions are placed in asserts.charInserts and consist of a length 
+/*Inserjions are placed in inserts.charInserts and consist of a length 
 * followed by the characters to be inserted. The length is one more than 
 * the numbr of characters to make it simpler to step through the 
 * insertions in the next function.*/
@@ -498,7 +517,7 @@ countAttrValues (xmlChar * key)
 	return 1;
       if (curCount >= NUMCOUNTS)
 	return 0;
-      hashInsert (attrValueCountsTable, key, curCount, NULL, NULL);
+      hashInsert (attrValueCountsTable, key, 0, curCount, NULL, NULL);
       curCount++;
       return 1;
     case 3:
@@ -509,7 +528,7 @@ countAttrValues (xmlChar * key)
       if (thisCount == notFound)
 	{
 	  attrValueCounts[curCount]++;
-	  hashInsert (attrValueCountsTable, key, curCount, NULL, NULL);
+	  hashInsert (attrValueCountsTable, key, 0, curCount, NULL, NULL);
 	  curCount++;
 	}
       key[lastComma] = ',';
@@ -536,11 +555,31 @@ destroyattrValueCountsTable (void)
 }
 static int sem_compileFile (const char *fileName);
 
+int
+find_group_length (const char groupSym[2], const char *groupStart)
+{
+  int level = 0;
+  int k;
+  if (*groupStart != groupSym[0])
+    return -1;
+  for (k = 0; groupStart[k]; k++)
+    {
+      if (groupStart[k] == groupSym[0])
+	level++;
+      if (groupStart[k] == groupSym[1])
+	level--;
+      if (level == 0)
+	return k + 1;
+    }
+  return -1;
+}
 static int
 compileLine (FileInfo * nested)
 {
   char *curchar = NULL;
   int ch = 0;
+  int func = 0;
+  EntryType type = 0;
   char *action = NULL;
   int actionLength = 0;
   char *lookFor;
@@ -569,8 +608,60 @@ compileLine (FileInfo * nested)
       return 0;
     }
   lookFor = curchar - 1;
-  while (*curchar++ > 32);
-  lookForLength = curchar - lookFor - 1;
+  if (*lookFor == '&')
+    {
+      /*xpath or other special case */
+      static const char *funcNames[] = {
+	"xpath",
+	"1",
+	NULL
+      };
+      char *funcName;
+      int funcNameLength;
+      char *argsStart;
+      int argsLength;
+      while ((ch = *curchar++) <= 32 && ch != 0);
+      funcName = curchar - 1;
+      while ((ch = *curchar++) > 32 && ch != '(');
+      funcNameLength = curchar - funcName - 1;
+      funcName[funcNameLength] = 0;
+      func = find_action (funcNames, funcName);
+      if (func < 0)
+	{
+	  semanticError (nested,
+			 "function name '%s' in column 2 not recognized",
+			 funcName);
+	  return 0;
+	}
+      funcName[funcNameLength] = ch;
+      if (ch != '(')
+	while ((ch = *curchar++) <= 32 && ch != 0);
+      argsStart = curchar - 1;
+      argsLength = find_group_length ("()", argsStart);
+      if (argsLength < 0)
+	{
+	  semanticError (nested, "unmatched parentheses in column 2 '%s'",
+			 lookFor);
+	  return 0;
+	}
+      switch (func)
+	{
+	case 1:
+	  type |= xpathEntry;
+	  break;
+	default:
+	  break;
+	}
+      lookFor = argsStart;
+      *lookFor = '&';
+      lookForLength = argsLength - 1;
+      curchar = &lookFor[argsLength];
+    }
+  else
+    {
+      while (*curchar++ > 32);
+      lookForLength = curchar - lookFor - 1;
+    }
   lookFor[lookForLength] = 0;
   actionNum = find_semantic_number (action);
   style = lookup_style (action);
@@ -592,13 +683,20 @@ compileLine (FileInfo * nested)
 	case end_all + 2:	/*newentries */
 	  docNewEntries = 0;
 	  break;
+	case end_all + 3:
+	  registerNamespaces (nested, xpathCtx, lookFor);
+	  break;
 	default:
 	  break;
 	}
       return 1;
     }
   if (hashLookup (semanticTable, (xmlChar *) lookFor) != notFound)
-    return 1;
+    {
+      if (ud->debug)
+	semanticError (nested, "duplicate entry '%s' in column 2", lookFor);
+      return 1;
+    }
   countAttrValues ((xmlChar *) lookFor);
   inserts = NULL;
   while ((ch = *curchar++) <= 32 && ch != 0);
@@ -619,7 +717,8 @@ compileLine (FileInfo * nested)
     }
   if (actionNum < 0)
     actionNum = generic;
-  hashInsert (semanticTable, (xmlChar *) lookFor, actionNum, inserts, style);
+  hashInsert (semanticTable, (xmlChar *) lookFor, type, actionNum,
+	      inserts, style);
   nested->numEntries++;
   return 1;
 }
@@ -714,7 +813,7 @@ getRootName (xmlNode * rootElement, char *fileName)
   curchar = fileName;
   while (*curchar)
     {
-      if (*curchar == ':' || *curchar == '/'
+      if (*curchar == ':' || *curchar == '/' || *curchar == ','
 	  || *curchar == '\\' || *curchar == 34 || *curchar == 39
 	  || *curchar == '(' || *curchar == ')' || *curchar < 33 ||
 	  *curchar > 126)
@@ -736,14 +835,14 @@ compile_semantic_table (xmlNode * rootElement)
   numEntries = 0;
   if (*ud->semantic_files)
     {
-/*Process file list*/
+      /*Process file list */
       int listLength;
       int currentListPos;
       int k;
       listLength = strlen (ud->semantic_files);
-      if (strcmp (ud->semantic_files, oldFileList) == 0)
+      if (ud->mode & dontInit)
 	return 1;
-      strcpy (oldFileList, ud->semantic_files);
+      xpathCtx = xmlXPathNewContext (rootElement->doc);
       firstFileName[0] = 0;
       for (k = 0; k < listLength; k++)
 	if (ud->semantic_files[k] == ',')
@@ -787,9 +886,9 @@ compile_semantic_table (xmlNode * rootElement)
   else
     {
       getRootName (rootElement, fileName);
-      if (strcmp (fileName, oldFileList) == 0)
+    if (ud->mode != 0)
 	return 1;
-      strcpy (oldFileList, fileName);
+      xpathCtx = xmlXPathNewContext (rootElement->doc);
       strcpy (firstFileName, fileName);
       if (!sem_compileFile (fileName))
 	return (haveSemanticFile = 0);
@@ -809,6 +908,148 @@ compile_semantic_table (xmlNode * rootElement)
 
 static void addNewEntries (const xmlChar * key);
 
+/**
+ * registerNamespaces:
+ * @xpathCtx:		the pointer to an XPath context.
+ * @nsList:		the list of known namespaces in 
+ *			"<prefix1>=<href1> <prefix2>=href2> ..." format.
+ *
+ * Registers namespaces from @nsList in @xpathCtx.
+ *
+ * Returns 1 on success and 0 on failure.
+ */
+static int
+registerNamespaces (FileInfo * nested, xmlXPathContextPtr xpathCtx, const
+		    xmlChar * nsList)
+{
+  xmlChar *nsListDup;
+  xmlChar *prefix;
+  xmlChar *href;
+  xmlChar *next;
+
+  nsListDup = xmlStrdup (nsList);
+  if (nsListDup == NULL)
+    {
+      semanticError (nested, "Error: unable to strdup namespaces list");
+      return 0;
+    }
+  next = nsListDup;
+  while (next != NULL)
+    {
+      if ((*next) == '\0')
+	break;
+      /* find prefix */
+      prefix = next;
+      next = (xmlChar *) xmlStrchr (next, '=');
+      if (next == NULL)
+	{
+	  semanticError (nested, "Error: invalid namespaces list format");
+	  xmlFree (nsListDup);
+	  return 0;
+	}
+      *(next++) = '\0';
+      /* find href */
+      href = next;
+      next = (xmlChar *) xmlStrchr (next, ',');
+      if (next != NULL)
+	{
+	  *(next++) = '\0';
+	}
+      /* do register namespace */
+      if (xmlXPathRegisterNs (xpathCtx, prefix, href) != 0)
+	{
+	  semanticError (nested,
+			 "Error: unable to register NS with prefix=\"%s\" and href=\"%s\"",
+			 prefix, href);
+	  xmlFree (nsListDup);
+	  return 0;
+	}
+    }
+  xmlFree (nsListDup);
+  return 1;
+}
+
+static void
+printXpathNodes (xmlNodeSetPtr nodes)
+{
+  xmlNodePtr cur;
+  int size;
+  int i;
+  size = (nodes) ? nodes->nodeNr : 0;
+  semanticError (NULL, "Result (%d nodes):", size);
+  for (i = 0; i < size; ++i)
+    {
+      if (nodes->nodeTab[i]->type == XML_NAMESPACE_DECL)
+	{
+	  xmlNsPtr ns;
+	  ns = (xmlNsPtr) nodes->nodeTab[i];
+	  cur = (xmlNodePtr) ns->next;
+	  if (cur->ns)
+	    {
+	      semanticError (NULL,
+			     "= namespace \"%s\"=\"%s\" for node %s:%s",
+			     ns->prefix, ns->href, cur->ns->href, cur->name);
+	    }
+	  else
+	    {
+	      semanticError (NULL, "= namespace \"%s\"=\"%s\" for node %s",
+			     ns->prefix, ns->href, cur->name);
+	    }
+	}
+      else if (nodes->nodeTab[i]->type == XML_ELEMENT_NODE)
+	{
+	  cur = nodes->nodeTab[i];
+	  if (cur->ns)
+	    {
+	      semanticError (NULL, "= element node \"%s:%s\"",
+			     cur->ns->href, cur->name);
+	    }
+	  else
+	    {
+	      semanticError (NULL, "= element node \"%s\"", cur->name);
+	    }
+	}
+      else
+	{
+	  cur = nodes->nodeTab[i];
+	  semanticError (NULL, "= node \"%s\": type %d", cur->name,
+			 cur->type);
+	}
+    }
+}
+
+int
+do_xpath_expr ()
+{
+  xmlXPathObject *xpathObj;
+  HashEntry *curEntry;
+  while ((curEntry = hashScan (semanticTable)))
+    {
+      if (curEntry->type & xpathEntry)
+	{
+	  xmlNodeSet *nodeSet;
+	  xmlNode *node;
+	  int size;
+	  int k;
+	  xpathObj = xmlXPathEvalExpression (&curEntry->key[1], xpathCtx);
+	  if (xpathObj == NULL || xpathObj->type != XPATH_NODESET)
+	    continue;
+	  nodeSet = xpathObj->nodesetval;
+	  if (ud->debug)
+	    printXpathNodes (nodeSet);
+	  size = (nodeSet) ? nodeSet->nodeNr : 0;
+	  for (k = 0; k < size; k++)
+	    {
+	      node = nodeSet->nodeTab[k];
+	      if (node->_private == NULL)
+		node->_private = curEntry;
+	    }
+	  xmlXPathFreeObject (xpathObj);
+	}
+    }
+  return 1;
+}
+
 sem_act
 set_sem_attr (xmlNode * node)
 {
@@ -818,6 +1059,8 @@ set_sem_attr (xmlNode * node)
   int k;
   int oldKeyLength = 0;
   const xmlChar *name;
+  if (node->_private != NULL)
+    return get_sem_attr (node);
   if (node->type == XML_CDATA_SECTION_NODE)
     name = (xmlChar *) "cdata-section";
   else
@@ -881,7 +1124,7 @@ get_sem_attr (xmlNode * node)
 {
   HashEntry *nodeEntry = (HashEntry *) node->_private;
   if (nodeEntry != NULL)
-    return nodeEntry->value;
+    return nodeEntry->semNum;
   else
     return no;
 }
@@ -958,7 +1201,7 @@ addNewEntries (const xmlChar * newEntry)
     }
   if (hashLookup (newEntriesTable, newEntry) != notFound)
     return;
-  hashInsert (newEntriesTable, newEntry, 0, NULL, NULL);
+  hashInsert (newEntriesTable, newEntry, 0, 0, NULL, NULL);
 }
 
 void
@@ -1058,7 +1301,7 @@ new_style (xmlChar * name)
     return latestEntry->style;
   style = malloc (sizeof (StyleType));
   memset (style, 0, sizeof (StyleType));
-  hashInsert (semanticTable, key, -100, NULL, style);
+  hashInsert (semanticTable, key, styleEntry, 0, NULL, style);
   return style;
 }
 
