@@ -36,12 +36,13 @@
 
 #define LETSIGN "\\_"
 static StyleRecord *styleSpec;
-static StyleRecord *prevStyleSpec;
+/* Note that the following is an actual data area, not a pointer*/
+static StyleRecord prevStyleSpec;
 static StyleType *style;
 static StyleType *prevStyle;
 static int styleBody (void);
 
-int
+static int
 fineFormat (void)
 {
   if (ud->text_length == 0 && ud->translated_length == 0)
@@ -90,8 +91,9 @@ start_document (void)
   else
     ud->braille_page_number = ud->beginning_braille_page_number;
   ud->outlen_so_far = 0;
-  styleSpec = prevStyleSpec = NULL;
+  styleSpec = &prevStyleSpec;
   style = prevStyle = lookup_style ("document");
+  prevStyleSpec.style = prevStyle;
   if (ud->outFile && ud->output_encoding == utf16)
     {
       /*Little Endian indicator */
@@ -121,6 +123,8 @@ start_document (void)
 int
 end_document (void)
 {
+  if (ud->style_top < 0)
+   ud->style_top = 0;
   if (ud->text_length != 0)
     insert_translation (ud->mainBrailleTable);
   if (ud->translated_length != 0)
@@ -1043,7 +1047,6 @@ writeOutbuf (void)
 {
   int k;
   unsigned char *utf8Str;
-  ud->translated_length = 0;
   if (ud->outlen_so_far == 0 || ud->outFile == NULL)
     return 1;			/*output stays in ud->outbuf */
   switch (ud->output_encoding)
@@ -1616,65 +1619,57 @@ doLeftJustify (void)
 static int
 doContents (void)
 {
-  int orig_translatedLength = 0;
+  int lastWord;
+  int lastWordLength;
+  int untilLastWord;
+  int numbersStart;
+  int numbersLength;
+  int leadingBlanks = 0;
   int charactersWritten = 0;
   int cellsToWrite = 0;
   int availableCells = 0;
   int k;
-  int numbersLength = 0;
-  int sameLine = 0;
-  if ((!ud->braille_pages && !ud->print_pages) || ud->line_fill == ' ')
+  if (translatedBuffer[translatedLength - 1] == 0xa0)
     {
+      /* No page numbers */
+      translatedLength--;
       doLeftJustify ();
       return 1;
     }
-  orig_translatedLength = translatedLength;
   for (k = translatedLength - 1; k > 0 && translatedBuffer[k] != 32; k--);
   if (k == 0)
     {
       doLeftJustify ();
       return 1;
     }
-  translatedLength = k;
-  while (charactersWritten < translatedLength)
+  numbersStart = k + 1;
+  numbersLength = translatedLength - numbersStart;
+  for (--k; k >= 0 && translatedBuffer[k] > 32; k--);
+  lastWord = k + 1;
+  lastWordLength = numbersStart - lastWord;
+  for (k = numbersStart; k < translatedLength; k++)
+    if (translatedBuffer[k] == 0xa0)
+      translatedBuffer[k] = ' ';
+  untilLastWord = lastWord - 1;
+  while (charactersWritten < untilLastWord)
     {
       int wordTooLong = 0;
       int breakAt = 0;
-      int leadingBlanks = 0;
-int lastLineCells;
-      if (!sameLine)
+      availableCells = startLine ();
+      if (styleSpec->status == startBody)
 	{
-	  availableCells = startLine ();
-	  if (styleSpec->status == startBody)
-	    {
-	      leadingBlanks = style->left_margin + style->first_line_indent;
-	      styleSpec->status = resumeBody;
-	    }
-	  else
-	    leadingBlanks = style->left_margin;
-if (leadingBlanks < 0)
-leadingBlanks = 0;
-	  if (!insertCharacters (blanks, leadingBlanks))
-	    return 0;
-	  availableCells -= leadingBlanks;
-lastLineCells = availableCells - (orig_translatedLength - 
-translatedLength - 2);
+	  leadingBlanks = style->left_margin + style->first_line_indent;
+	  styleSpec->status = resumeBody;
 	}
-      if (numbersLength)
-	{
-	  if ((availableCells - numbersLength) <= 3)
-	    insertCharacters (blanks, availableCells - numbersLength);
-	  else
-	    {
-	      insertCharacters (blanks, 1);
-	      for (k = availableCells - (numbersLength + 1); k > 0; k--)
-		insertCharacters (&ud->line_fill, 1);
-	      insertCharacters (blanks, 1);
-	    }
-	  availableCells = numbersLength;
-	}
-      if ((charactersWritten + availableCells) >= translatedLength)
-	cellsToWrite = translatedLength - charactersWritten;
+      else
+	leadingBlanks = style->left_margin;
+      if (leadingBlanks < 0)
+	leadingBlanks = 0;
+      if (!insertCharacters (blanks, leadingBlanks))
+	return 0;
+      availableCells -= leadingBlanks;
+      if ((charactersWritten + availableCells) >= untilLastWord)
+	cellsToWrite = untilLastWord - charactersWritten;
       else
 	{
 	  for (cellsToWrite = availableCells - 2; cellsToWrite > 0;
@@ -1711,19 +1706,76 @@ translatedLength - 2);
 	  if (!insertDubChars (litHyphen, strlen (litHyphen)))
 	    return 0;
 	}
-      if (charactersWritten >= translatedLength && !numbersLength)
+      if (charactersWritten < untilLastWord)
+	finishLine ();
+      else
 	{
-	  numbersLength = orig_translatedLength - translatedLength;
-	  translatedLength = orig_translatedLength;
-	  if ((availableCells - cellsToWrite) < (numbersLength + 2))
-	    finishLine ();
-	  else
+	  availableCells -= cellsToWrite;
+	  if (availableCells <= 0)
 	    {
-	      availableCells -= cellsToWrite;
-	      sameLine = 1;
+	      finishLine ();
+	      availableCells = 0;
 	    }
-	  continue;
 	}
+    }
+  if (availableCells == 0)
+    {
+      availableCells = startLine ();
+      if (styleSpec->status == startBody)
+	{
+	  leadingBlanks = style->left_margin + style->first_line_indent;
+	  styleSpec->status = resumeBody;
+	}
+      else
+	leadingBlanks = style->left_margin;
+      if (leadingBlanks < 0)
+	leadingBlanks = 0;
+      if (!insertCharacters (blanks, leadingBlanks))
+	return 0;
+      availableCells -= leadingBlanks;
+    }
+  if ((lastWordLength + numbersLength + 2) < availableCells)
+    {
+      insertCharacters (blanks, 1);
+      availableCells--;
+      if (!insertWidechars (&translatedBuffer[lastWord], lastWordLength))
+	return 0;
+      availableCells -= lastWordLength;
+      if ((availableCells - numbersLength) < 3)
+	insertCharacters (blanks, availableCells - numbersLength);
+      else
+	{
+	  insertCharacters (blanks, 1);
+	  for (k = availableCells - (numbersLength + 1); k > 0; k--)
+	    insertCharacters (&ud->line_fill, 1);
+	  insertCharacters (blanks, 1);
+	}
+      if (!insertWidechars (&translatedBuffer[numbersStart], numbersLength))
+	return 0;
+      finishLine ();
+    }
+  else
+    {
+      finishLine ();
+      availableCells = startLine ();
+      leadingBlanks = style->left_margin;
+      if (!insertCharacters (blanks, leadingBlanks))
+	return 0;
+      availableCells -= leadingBlanks;
+      if (!insertWidechars (&translatedBuffer[lastWord], lastWordLength))
+	return 0;
+      availableCells -= lastWordLength;
+      if ((availableCells - numbersLength) < 3)
+	insertCharacters (blanks, availableCells - numbersLength);
+      else
+	{
+	  insertCharacters (blanks, 1);
+	  for (k = availableCells - (numbersLength + 1); k > 0; k--)
+	    insertCharacters (&ud->line_fill, 1);
+	  insertCharacters (blanks, 1);
+	}
+      if (!insertWidechars (&translatedBuffer[numbersStart], numbersLength))
+	return 0;
       finishLine ();
     }
   return 1;
@@ -1868,7 +1920,7 @@ styleBody (void)
   while (ud->translated_length > 0 &&
 	 ud->translated_buffer[ud->translated_length - 1] <= 32)
     ud->translated_length--;
-  if (ud->translated_length == 0 || action == no)
+  if (ud->translated_length == 0)
     return 1;
   if (!editTrans ())
     return 0;
@@ -1889,7 +1941,10 @@ styleBody (void)
 	 translatedBuffer[translatedLength - 1] != escapeChar)
     translatedLength--;
   if (translatedLength <= 0)
-    return 1;
+    {
+      ud->translated_length = 0;
+      return 1;
+    }
   if (!ud->paragraphs)
     {
       cellsWritten = 0;
@@ -1903,6 +1958,7 @@ styleBody (void)
       else if (!insertCharacters (ud->lineEnd, strlen (ud->lineEnd)))
 	return 0;
       writeOutbuf ();
+      ud->translated_length = 0;
       return 1;
     }
   if (action == contentsheader && ud->contents != 2)
@@ -1958,6 +2014,7 @@ styleBody (void)
   if (ud->contents == 1)
     finish_heading (action);
   styleSpec->status = resumeBody;
+  ud->translated_length = 0;
   return 1;
 }
 
@@ -1997,11 +2054,22 @@ write_paragraph (sem_act action)
 {
   StyleType *holdStyle = action_to_style (action);
   if (!((ud->text_length > 0 || ud->translated_length > 0) &&
-      ud->style_top >= 0))
+	ud->style_top >= 0))
     return 1;
   if (holdStyle == NULL)
     holdStyle = lookup_style ("para");
-  start_style (holdStyle);
+  /* We must do some of the work of start_styl* */
+  if (ud->style_top < (STACKSIZE - 2))
+    ud->style_top++;
+  styleSpec = &ud->style_stack[ud->style_top];
+  style = styleSpec->style = holdStyle;
+  styleSpec->status = beforeBody;
+  if (style->brlNumFormat != normal)
+    ud->brl_page_num_format = style->brlNumFormat;
+  styleSpec->curBrlNumFormat = ud->brl_page_num_format;
+  startStyle ();
+  insert_translation (ud->mainBrailleTable);
+  styleBody ();
   end_style (holdStyle);
   return 1;
 }
@@ -2454,17 +2522,7 @@ start_style (StyleType * curStyle)
   if (curStyle == NULL)
     curStyle = lookup_style ("para");
   if (prevStyle == NULL)
-    prevStyle = lookup_style ("document");
-  if (prevStyleSpec == NULL)
-    {
-      prevStyleSpec = &ud->style_stack[STACKSIZE - 1];
-      prevStyleSpec->style = curStyle;
-    }
-  else
-    {
-      prevStyleSpec = styleSpec;
-      prevStyle = style;
-    }
+    prevStyle = lookup_style ("para");
   if ((ud->text_length > 0 || ud->translated_length > 0) &&
       ud->style_top >= 0)
     {
@@ -2499,6 +2557,8 @@ end_style (StyleType * curStyle)
   insert_translation (ud->mainBrailleTable);
   styleBody ();
   finishStyle ();
+  memcpy (&prevStyleSpec, styleSpec, sizeof (prevStyleSpec));
+  prevStyle = prevStyleSpec.style;
   ud->style_top--;
   if (ud->style_top < 0)
     ud->style_top = 0;
